//	<><><><><><><><><><><><><>  PlotCalc.cpp  <><><><><><><><><><><><><><> 
//
// -----------------------------------------------------------------
// (C) Copyright Clark Dailey, Inc. 2006
//
// PlotCalc is a calculator class for user definable plotting functions
//
// -----------------------------------------------------------------

#include "stdafx.h"
#include "stdlib.h"
#include "float.h"
#include "plot3d.h"
#include "plotcalc.h"

// Prototyping
double Factorial(int n);
double Gamma(double x);
double LogGamma(double x);
double Psi(double x);
double Zetar(double real, double imag);
double Zetai(double real, double imag);
double Zetax(double real, double imag);

// --------------------------------------------------------------
// Supported Instructions
static const PcInstr g_Inst[] =
{
	// CAUTION! this order needs to match the order of PcOpCode enum
	{ PC_OP_CONSTANT,     PC_OP_OPERAND,      0,       0 }, // not used (padding for one based)
	{ PC_OP_OPEN_PAREN,   PC_OP_CONTROL,      "(",     0 },
	{ PC_OP_CLOSE_PAREN,  PC_OP_CONTROL,      ")",     0 },
	{ PC_OP_COMMA,        PC_OP_CONTROL,      ",",     0 },
	{ PC_OP_ADD,          PC_OP_OPERATOR,     "+",     0 },
	{ PC_OP_SUB,          PC_OP_OPERATOR,     "-",     0 },
	{ PC_OP_MULT,         PC_OP_OPERATOR,     "*",     0 },
	{ PC_OP_DIV,          PC_OP_OPERATOR,     "/",     0 },
	{ PC_OP_POW,          PC_OP_OPERATOR,     "^",     0 },
	{ PC_OP_POW2,         PC_OP_INTERNAL,     "^2",    0 }, // optimization
	{ PC_OP_POW3,         PC_OP_INTERNAL,     "^3",    0 }, // optimization
	{ PC_OP_POW4,         PC_OP_INTERNAL,     "^4",    0 }, // optimization
	{ PC_OP_NEG,          PC_OP_INTERNAL,     "neg",   0 },
	{ PC_OP_PI,           PC_OP_OPERAND,      "pi",    0 },
	{ PC_OP_E,            PC_OP_OPERAND,      "e",     0 },
    { PC_OP_X,            PC_OP_OPERAND,      "x",     0 },
    { PC_OP_Y,            PC_OP_OPERAND,      "y",     0 },
    { PC_OP_Z,            PC_OP_OPERAND,      "z",     0 },
    { PC_OP_U,            PC_OP_OPERAND,      "u",     0 }, // an alias for x
    { PC_OP_V,            PC_OP_OPERAND,      "v",     0 }, // an alias for y
    { PC_OP_C1,           PC_OP_OPERAND,      "c1",    0 },	// c1-c8 user definable constants
    { PC_OP_C2,           PC_OP_OPERAND,      "c2",    0 },
    { PC_OP_C3,           PC_OP_OPERAND,      "c3",    0 },
    { PC_OP_C4,           PC_OP_OPERAND,      "c4",    0 },
    { PC_OP_C5,           PC_OP_OPERAND,      "c5",    0 },
    { PC_OP_C6,           PC_OP_OPERAND,      "c6",    0 },
    { PC_OP_C7,           PC_OP_OPERAND,      "c7",    0 },
    { PC_OP_C8,           PC_OP_OPERAND,      "c8",    0 },
    { PC_OP_SETC1,        PC_OP_INTERNAL,     "c1=",   0 },	// set c1-c8 user constants
    { PC_OP_SETC2,        PC_OP_INTERNAL,     "c2=",   0 },
    { PC_OP_SETC3,        PC_OP_INTERNAL,     "c3=",   0 },
    { PC_OP_SETC4,        PC_OP_INTERNAL,     "c4=",   0 },
    { PC_OP_SETC5,        PC_OP_INTERNAL,     "c5=",   0 },
    { PC_OP_SETC6,        PC_OP_INTERNAL,     "c6=",   0 },
    { PC_OP_SETC7,        PC_OP_INTERNAL,     "c7=",   0 },
    { PC_OP_SETC8,        PC_OP_INTERNAL,     "c8=",   0 },
    { PC_OP_SQRT,         PC_OP_FUNCTION,     "sqrt",  1 },
    { PC_OP_ABS,          PC_OP_FUNCTION,     "abs",   1 },
    { PC_OP_SIN,          PC_OP_FUNCTION,     "sin",   1 },
    { PC_OP_COS,          PC_OP_FUNCTION,     "cos",   1 },
    { PC_OP_TAN,          PC_OP_FUNCTION,     "tan",   1 },
    { PC_OP_ASIN,         PC_OP_FUNCTION,     "asin",  1 },
    { PC_OP_ACOS,         PC_OP_FUNCTION,     "acos",  1 },
    { PC_OP_ATAN,         PC_OP_FUNCTION,     "atan",  1 },
    { PC_OP_SINH,         PC_OP_FUNCTION,     "sinh",  1 },
    { PC_OP_COSH,         PC_OP_FUNCTION,     "cosh",  1 },
    { PC_OP_TANH,         PC_OP_FUNCTION,     "tanh",  1 },
    { PC_OP_EXP,          PC_OP_FUNCTION,     "exp",   1 },
    { PC_OP_LN,           PC_OP_FUNCTION,     "ln",    1 },
    { PC_OP_LOG10,        PC_OP_FUNCTION,     "log10", 1 },
    { PC_OP_CEIL,         PC_OP_FUNCTION,     "ceil",  1 },
    { PC_OP_FLOOR,        PC_OP_FUNCTION,     "floor", 1 },
    { PC_OP_RAND,         PC_OP_FUNCTION,     "rand",  0 },
    { PC_OP_MIN,          PC_OP_FUNCTION,     "min",   2 },
    { PC_OP_MAX,          PC_OP_FUNCTION,     "max",   2 },
    { PC_OP_ATAN2,        PC_OP_FUNCTION,     "atan2", 2 },
    { PC_OP_HYPOT,        PC_OP_FUNCTION,     "hypot", 2 },
    { PC_OP_SIGN,         PC_OP_FUNCTION,     "sign",  1 },

    // ### new functions added
    { PC_OP_FACTORIAL,    PC_OP_FUNCTION,     "nfact", 1 },
    { PC_OP_IF_LT,        PC_OP_FUNCTION,     "if_lt", 4 },
    { PC_OP_IF_LE,        PC_OP_FUNCTION,     "if_le", 4 },
    { PC_OP_IF_GT,        PC_OP_FUNCTION,     "if_gt", 4 },
    { PC_OP_IF_GE,        PC_OP_FUNCTION,     "if_ge", 4 },
    { PC_OP_FMOD,         PC_OP_FUNCTION,     "fmod",  2 },
    { PC_OP_ASINH,        PC_OP_FUNCTION,     "asinh", 1 },
    { PC_OP_ACOSH,        PC_OP_FUNCTION,     "acosh", 1 },
    { PC_OP_ATANH,        PC_OP_FUNCTION,     "atanh", 1 },

    { PC_OP_J0,           PC_OP_FUNCTION,     "j0",    1 },
    { PC_OP_J1,           PC_OP_FUNCTION,     "j1",    1 },
    { PC_OP_JN,           PC_OP_FUNCTION,     "jn",    2 },
    { PC_OP_Y0,           PC_OP_FUNCTION,     "y0",    1 },
    { PC_OP_Y1,           PC_OP_FUNCTION,     "y1",    1 },
    { PC_OP_YN,           PC_OP_FUNCTION,     "yn",    2 },

    { PC_OP_GAMMA,        PC_OP_FUNCTION,     "gamma", 1 },
    { PC_OP_LGAMMA,       PC_OP_FUNCTION,     "lgamma",1 },
    { PC_OP_PSI,          PC_OP_FUNCTION,     "psi",   1 },
    { PC_OP_ZETA,         PC_OP_FUNCTION,     "zeta",  2 },
    { PC_OP_ZETAR,        PC_OP_FUNCTION,     "zetar", 2 },
    { PC_OP_ZETAI,        PC_OP_FUNCTION,     "zetai", 2 },
};

#define	TOTAL_OP_CODES	(sizeof(g_Inst)/sizeof(g_Inst[0]))

#define	CONSTANT_TOLERANCE	(1.e-10) // tolerance allowed for constant equality

// add program instruction
// returns: 0=ok, 1=too many instructions
int PlotCalc::AddInst(PcOpCode inst,unsigned char operand)
{
	double c;
	if (m_nInst>=PC_MAX_INSTRUCTIONS)
		return(1);
	// toss any non-executable instructions
	if (inst==PC_OP_OPEN_PAREN || inst==PC_OP_CLOSE_PAREN || inst==PC_OP_COMMA) return(0);
	if (inst==PC_OP_POW)
	{
		// optimize for ^2, ^3, ^4
		if (m_nInst && m_opcodes[m_nInst-1]==PC_OP_CONSTANT)
		{
			c = m_dconstants[m_operands[m_nInst-1]];
			if (fabs(c-2.)<CONSTANT_TOLERANCE)
			{
				m_opcodes[ m_nInst-1] = (unsigned char)PC_OP_POW2;
				m_operands[m_nInst-1] = 0;
				return(0);
			}
			if (fabs(c-3.)<CONSTANT_TOLERANCE)
			{
				m_opcodes[ m_nInst-1] = (unsigned char)PC_OP_POW3;
				m_operands[m_nInst-1] = 0;
				return(0);
			}
			if (fabs(c-4.)<CONSTANT_TOLERANCE)
			{
				m_opcodes[ m_nInst-1] = (unsigned char)PC_OP_POW4;
				m_operands[m_nInst-1] = 0;
				return(0);
			}
		}
	}
	if (inst>=PC_OP_SETC1 && inst<=PC_OP_SETC8)
	{
		// optimize for constant user constant
		if (m_nInst && m_opcodes[m_nInst-1]==PC_OP_CONSTANT)
		{
			switch(inst)
			{
			case PC_OP_SETC1: m_c1 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC2: m_c2 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC3: m_c3 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC4: m_c4 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC5: m_c5 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC6: m_c6 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC7: m_c7 = m_dconstants[m_operands[m_nInst-1]]; break;
			case PC_OP_SETC8: m_c8 = m_dconstants[m_operands[m_nInst-1]]; break;
			}
			m_nInst--;
			return(0);
		}
	}
	m_opcodes[ m_nInst] = (unsigned char)inst;
	m_operands[m_nInst] = operand;
	m_nInst++;
	return(0);
}

// --------------------------------------------------------------
int OpIndex(PcOpCode opcode)
{
	unsigned i;
	for (i=0; i<TOTAL_OP_CODES; i++)
	{
		if (opcode == g_Inst[i].opcode) return(i);
	}
	return(0);
}

// --------------------------------------------------------------
// Load program
// returns: 0=ok, else pointer to error in string
int PlotCalc::LoadProgram(char* label,const char* pgm,int allow_z,
		const char* pc1,const char* pc2,const char* pc3,const char* pc4,
		const char* pc5,const char* pc6,const char* pc7,const char* pc8,int use_uv)
{
	PcOpCode this_opcode, last_opcode;
	PcOpCode opcode2, subcode;
	int  i, len, prec, prec_tos, in_table;
	int  it, is_done, is_first;
	unsigned int  iexp, next_con=1; // constants counter
	char byte, token[100];
	char* pCode [PC_NO_USER_CONSTANTS+1];
	char* pCode0[PC_NO_USER_CONSTANTS+1];
	char errbuf[128];
	double dval;

	// assume no error
	sprintf(m_sErr,"Ok");

	// validate function arguments
	if (pgm==NULL ||
		pc1==NULL || pc2==NULL || pc3==NULL || pc4==NULL ||
		pc5==NULL || pc6==NULL || pc7==NULL || pc8==NULL)
	{
		strcpy(m_sErr,"LoadProgram() called with null pointer");
		return(1);
	}

	// if using u,v don't allow z coordinate
	if (use_uv)
		allow_z = 0;

	// clear program
	NewProgram();

	// default constants
	m_c1 = 1.;	m_c2 = 1.;	m_c3 = 1.;	m_c4 = 1.;
	m_c5 = 1.;	m_c6 = 1.;	m_c7 = 1.;	m_c8 = 1.;

	// setup pointers
	pCode[0] = pCode0[0] = (char*)pc1;
	pCode[1] = pCode0[1] = (char*)pc2;
	pCode[2] = pCode0[2] = (char*)pc3;
	pCode[3] = pCode0[3] = (char*)pc4;
	pCode[4] = pCode0[4] = (char*)pc5;
	pCode[5] = pCode0[5] = (char*)pc6;
	pCode[6] = pCode0[6] = (char*)pc7;
	pCode[7] = pCode0[7] = (char*)pc8;
	pCode[8] = pCode0[8] = (char*)pgm;

	// --------------------
	// parse program string
	// --------------------

	for (iexp=0; iexp<PC_NO_USER_CONSTANTS+1; iexp++)
	{
		// init values
		it = 0;
		is_done = 0;
		is_first = 1;
		this_opcode = PC_OP_CONSTANT,
		last_opcode = PC_OP_CONSTANT;
	
		while (!is_done)
		{
			// get next byte
			byte = *pCode[iexp];
	
			// skip white space
			if (IsWhiteSpace(byte))
			{
				pCode[iexp]++;
				continue;
			}
	
			// is it a token separator
			if (!IsTokenSeparator(byte,&this_opcode))
			{
				// save the byte and keep looking for separator
				pCode[iexp]++;
				token[it] = byte;
				it++;
				if ((it+2) <= sizeof(token)) continue;
				strcpy(errbuf,"Unrecognized operand; too many characters");
				goto abort_load;
			}
	
			if (it==0)
			{
				// process the separator
				pCode[iexp]++;
				switch(this_opcode)
				{
				case PC_OP_END_INPUT:
					is_done = 1;
					break;
				case PC_OP_OPEN_PAREN:
					PushOp(PC_OP_OPEN_PAREN,PC_OP_CONSTANT);
					break;
				case PC_OP_CLOSE_PAREN:
					for (;;)
					{
						if (PopOp(&opcode2,&subcode) != 0)
						{
				 			strcpy(errbuf,"Missing opening parenthesis");
							goto abort_load;
						}
						if (opcode2 == PC_OP_OPEN_PAREN) break;
						if (opcode2 == PC_OP_START_FUNC)
						{
							if (m_got_operand) m_cur_parms++;
							if (m_cur_parms != m_req_parms)
							{
								i = OpIndex(subcode);
					 			sprintf(errbuf,"Function %s( ) requires %d parameter(s)",
									g_Inst[i].mnemonic ,(int)m_req_parms);
								goto abort_load;
							}
							AddInst(subcode,0);
							PopFunc(&m_cur_parms,&m_req_parms,&m_got_operand);
							break;
						}
						AddInst(opcode2,(unsigned char)subcode);
					} // for
					break;
				case PC_OP_COMMA:
					m_cur_parms++;
					if (m_req_parms<2 || m_cur_parms>m_req_parms || m_got_operand==0)
					{
						if (m_req_parms)
			 				strcpy(errbuf,"Too many function parameters");
						else
			 				strcpy(errbuf,"Improper use of comma");
		 				goto abort_load;
					}
					for (;;)
					{
						if (PeekOp(&opcode2,&subcode) != 0) break;
						if (opcode2 == PC_OP_COMMA) break;
						PopOp(&opcode2,&subcode);
						AddInst(opcode2,(unsigned char)subcode);
					} // for
					m_got_operand = 0; // don't have an operand yet
					break;
				case PC_OP_SUB:
					// check for negation
					if (is_first ||
					   (last_opcode==PC_OP_OPEN_PAREN ||
					    last_opcode==PC_OP_START_FUNC ||
						last_opcode==PC_OP_COMMA))
							this_opcode = PC_OP_NEG; // use negation operator
					// fall through with other operators
				case PC_OP_ADD:
				case PC_OP_MULT:
				case PC_OP_DIV:
				case PC_OP_POW:
					if (IsOperator(last_opcode))
					{
		 				strcpy(errbuf,"Two adjacent operators");
		 				goto abort_load;
					}
					// process operator
					prec = OpPrecedence(this_opcode);
					for (;;)
					{
						if (PeekOp(&opcode2,&subcode) != 0) break;
						prec_tos = OpPrecedence(opcode2);
						if (prec_tos < prec) break;
						PopOp(&opcode2,&subcode);
						AddInst(opcode2,0);
						if (prec_tos <= prec) break;
					}
					PushOp(this_opcode,PC_OP_CONSTANT);
					break;
				} // switch
			}
			else // it != 0
			{
				// Note: be sure to update this_opcode
	
				// process the token
				token[it] = 0; // null terminate
				len = strlen(token);
		
		 		// check if token is in table
		 		for (i=1, in_table=0; i<TOTAL_OP_CODES; i++)
		 		{
					if (g_Inst[i].optype == PC_OP_INTERNAL) continue; // ignore internal commands
		 			if (stricmp(g_Inst[i].mnemonic,token)!=0) continue;
					if ((allow_z==0) && (g_Inst[i].opcode==PC_OP_Z)) continue;
					if (use_uv)
					{
						// don't allow x,y when using u,v
						if (g_Inst[i].opcode==PC_OP_X) continue;
						if (g_Inst[i].opcode==PC_OP_Y) continue;
					}
					else
					{
						// don't allow u,v when using x,y
						if (g_Inst[i].opcode==PC_OP_U) continue;
						if (g_Inst[i].opcode==PC_OP_V) continue;
					}

					if (g_Inst[i].opcode>=PC_OP_C1 && g_Inst[i].opcode<=PC_OP_C8)
					{
						unsigned constant_no = (g_Inst[i].opcode - PC_OP_C1);
						if (iexp <= constant_no)
						{
							sprintf(errbuf,"C%d cannot use C%d",iexp+1,constant_no+1);
							goto abort_load;
						}
					}
					in_table = 1;
					if (g_Inst[i].optype == PC_OP_FUNCTION)
					{
						if (this_opcode != PC_OP_OPEN_PAREN)
						{
				 			strcpy(errbuf,"Missing open parenthesis for function call");
			 				goto abort_load;
						}
						m_got_operand = 1; // got an operand now
	
						// push onto function stack
						PushFunc(m_cur_parms,m_req_parms,m_got_operand);
						m_cur_parms = 0; // no parameters yet
						m_req_parms = g_Inst[i].nParms; // require this many
		
						// push open function call onto stack
						this_opcode = PC_OP_START_FUNC;
						PushOp(this_opcode, g_Inst[i].opcode);
						PushOp(PC_OP_COMMA, PC_OP_CONSTANT);
						m_got_operand = 0; // no operand yet
						pCode[iexp]++; // skip over the open parenthesis
					}
					else
					{
						if (g_Inst[i].optype == PC_OP_OPERAND) m_got_operand = 1;
						this_opcode = g_Inst[i].opcode;
						AddInst(this_opcode,0);
					}
					break;
				} // for
	
				// was it found in the lookup table?
				if (!in_table)
				{
			 		// not in lookup table; check if it is a number (constant)
			 		if (sscanf(token,"%lf",&dval) == 1)
			 		{
			 			if (SetConstant(next_con, dval) != 0)
			 			{
			 				strcpy(errbuf,"Too many constants");
			 				goto abort_load;
			 			}
						m_got_operand = 1; // got an operand now
						this_opcode = PC_OP_CONSTANT;
			 			AddInst(this_opcode,(unsigned char)next_con);
			 			next_con++;
			 		}
			 		else
			 		{
			 			strcpy(errbuf,"Unrecognized operand");
			 			goto abort_load;
			 		}
				} // if !in_table
	
				// reset token
				it = 0;
				token[0] = 0;
			} // else it != 0
	
			// update for next time
			last_opcode = this_opcode;
			is_first = 0;
	
		} // while !is_done
	
		// -----------------------------------
		//        input is done
		// pop and enqueue remaining operators
		// -----------------------------------
		for (;;)
		{
			if (PopOp(&this_opcode,&subcode) != 0) break;
			if (this_opcode==PC_OP_OPEN_PAREN || this_opcode==PC_OP_START_FUNC)
			{
	 			strcpy(errbuf,"Missing closing parenthesis");
				goto abort_load;
			}
			AddInst(this_opcode,0);
		}

		// set user constant
		if (iexp<PC_NO_USER_CONSTANTS)
		{
			this_opcode = (PcOpCode)(PC_OP_SETC1 + iexp);
			AddInst(this_opcode,0);
		}
	} // for iexp

	// ------------------------------------------------
	// Optimization: Eliminate unused constants C1...C8
	// ------------------------------------------------
	int ic,ip,is_cused,is_cset,first_inst,last_inst;
	PcOpCode op_cused, op_cset;

	// CAUTION! eliminate constants in reverse order (C8,C7...C1)
	// check each constant in reverse order
	for (ic=PC_NO_USER_CONSTANTS; ic>0; ic--)
	{
		// first if constant is used, set, and last instruction number
		op_cused = (PcOpCode)(PC_OP_C1   +ic-1);
		op_cset  = (PcOpCode)(PC_OP_SETC1+ic-1);
		is_cused = is_cset = last_inst = 0;
		for (ip=0; ip<(int)m_nInst; ip++)
		{
			if (m_opcodes[ip] == op_cused) is_cused=1;
			if (m_opcodes[ip] != op_cset ) continue;
			is_cset = 1;
			last_inst = ip;
		} // for ip
		if (is_cused || !is_cset) continue;
		// find first instruction number
		// search backwards until constant set instruction
		first_inst = 0;
		for (ip=last_inst-1; ip>=0; ip--)
		{
			if (m_opcodes[ip] < PC_OP_SETC1 ||
			    m_opcodes[ip] > PC_OP_SETC8 ) continue;
			first_inst = ip+1;
			break;
		} // for ip
		// delete the instruction(s)
		DelInsts(first_inst,last_inst);
	} // for ic
	return(0);

abort_load:
	NewProgram(); // clear program
	// calculate offset into string
	i = pCode[iexp] - pCode0[iexp];
	if (i==0) i=1;
	if (iexp==PC_NO_USER_CONSTANTS)
		sprintf(m_sErr,"[%s @ Col:%d] %s", label,i,errbuf);
	else
		sprintf(m_sErr,"[C%d @ Col:%d] %s", iexp+1,i,errbuf);
	return(i);

} // LoadProgram

// --------------------------------------------------------------
// get program instruction
// ip - instruction number to get (zero based)
// buf - buffer to received instruction into (dimension to 100)
// returns: 0=ok, 1=invalid instruction number
int  PlotCalc::GetProgramInst(unsigned ip,char* buf)
{
	unsigned i;

	if (ip>=m_nInst) return(1);
 	i = (unsigned)m_opcodes[ip];
 	if (i>=TOTAL_OP_CODES)
		{ sprintf(buf,"%04u: **Invalid** (%u)",ip,i); }
 	else if ((PcOpCode)i == PC_OP_CONSTANT)
		{ sprintf(buf,"%04u: %lf%c",ip,m_dconstants[m_operands[ip]],PC_END_OF_INSTRUCTION); }
 	else
		{ sprintf(buf,"%04u: %s%c",ip,g_Inst[i].mnemonic,PC_END_OF_INSTRUCTION); }
	return(0);
}

// --------------------------------------------------------------
// constructor
PlotCalc::PlotCalc()
{
	NewProgram();
	memset(m_sErr,0,sizeof(m_sErr));
}

// --------------------------------------------------------------
// initialize new program
void PlotCalc::NewProgram()
{
	m_sp = 0;
	m_opsp = 0;
	m_fsp = 0;
	m_nInst = 0;
	m_err_flags = 0;
	m_cur_parms = 0;
	m_req_parms = 0;
	m_got_operand = 0;
	memset(m_dstack,    0,sizeof(m_dstack    ));
	memset(m_dconstants,0,sizeof(m_dconstants));
	memset(m_opcodes,   0,sizeof(m_opcodes   ));
	memset(m_operands,  0,sizeof(m_operands  ));
	memset(m_opstack,   0,sizeof(m_opstack   ));
	memset(m_fstack,    0,sizeof(m_fstack    ));
}

// --------------------------------------------------------------
int PlotCalc::DelInsts(unsigned first_inst_no,unsigned last_inst_no)
{
	unsigned nmove;
	if (first_inst_no > last_inst_no) return(1);
	if (first_inst_no >= m_nInst) return(1);
	nmove = m_nInst - last_inst_no - 1;
	if (nmove)
	{
		memmove(&m_opcodes [first_inst_no],&m_opcodes [last_inst_no+1],nmove);
		memmove(&m_operands[first_inst_no],&m_operands[last_inst_no+1],nmove);
	}
	m_nInst -= (last_inst_no - first_inst_no + 1);
	return(0);
}

// --------------------------------------------------------------
// set a constant value
// number: 1-255, value=constant
// returns: 0=ok, 1=number out of range
int PlotCalc::SetConstant(unsigned number,double value)
{
	if (number<1 || number>PC_MAX_CONSTANTS)
		return(1);
	m_dconstants[number] = value;
	return(0);
}

// --------------------------------------------------------------
// returns: 0=ok, 1=stack overflow
int PlotCalc::PushOp(PcOpCode opcode, PcOpCode subcode)
{
	if (m_opsp>=PC_STACK_SIZE) return(1);
	m_opsp++;
	m_opstack[m_opsp][0] = opcode;
	m_opstack[m_opsp][1] = subcode;
	return(0);
}

// --------------------------------------------------------------
// returns: 0=ok, 1=stack underflow
int PlotCalc::PopOp(PcOpCode* opcode, PcOpCode* subcode)
{
	if (m_opsp<1) return(1);
	*opcode  = m_opstack[m_opsp][0];
	*subcode = m_opstack[m_opsp][1];
	m_opsp--;
	return(0);
}

// --------------------------------------------------------------
// returns: 0=ok, 1=stack underflow
int PlotCalc::PeekOp(PcOpCode* opcode, PcOpCode* subcode)
{
	if (m_opsp<1) return(1);
	*opcode  = m_opstack[m_opsp][0];
	*subcode = m_opstack[m_opsp][1];
	return(0);
}

// --------------------------------------------------------------
// returns: 0=ok, 1=stack overflow
int PlotCalc::PushFunc(unsigned char CurParms, unsigned char ReqParms, unsigned char GotOperand)
{
	if (m_fsp>=PC_STACK_SIZE) return(1);
	m_fsp++;
	m_fstack[m_fsp][0] = CurParms;
	m_fstack[m_fsp][1] = ReqParms;
	m_fstack[m_fsp][2] = GotOperand;
	return(0);
}

// --------------------------------------------------------------
// returns: 0=ok, 1=stack underflow
int PlotCalc::PopFunc(unsigned char* CurParms, unsigned char* ReqParms, unsigned char* GotOperand)
{
	if (m_fsp<1) return(1);
	*CurParms   = m_fstack[m_fsp][0];
	*ReqParms   = m_fstack[m_fsp][1];
	*GotOperand = m_fstack[m_fsp][2];
	m_fsp--;
	return(0);
}

// --------------------------------------------------------------
// determine operator precedence
int PlotCalc::OpPrecedence(PcOpCode opcode)
{
	int p=0;
	switch (opcode)
	{
	case PC_OP_START_FUNC:	p=1; break;
	case PC_OP_OPEN_PAREN:	p=1; break;
	case PC_OP_CLOSE_PAREN:	p=1; break;
	case PC_OP_COMMA	:	p=1; break;
	case PC_OP_ADD:			p=2; break;
	case PC_OP_SUB:			p=2; break;
	case PC_OP_MULT:		p=3; break;
	case PC_OP_DIV:			p=3; break;
	case PC_OP_POW:			p=4; break;
	case PC_OP_NEG:			p=5; break;
	} // switch
	return(p);
}

// --------------------------------------------------------------
// returns: 0=no, 1=it is an operator
int PlotCalc::IsOperator(PcOpCode opcode)
{
	if (opcode==PC_OP_ADD  ||
		opcode==PC_OP_SUB  ||
		opcode==PC_OP_MULT ||
		opcode==PC_OP_DIV  ||
		opcode==PC_OP_POW  ||
		opcode==PC_OP_NEG ) return(1);
	return(0);
}

// --------------------------------------------------------------
// determine if byte is a token separator
// returns: 0=no, 1=yes (and opcode is valid)
int PlotCalc::IsTokenSeparator(char byte,PcOpCode* pCode)
{
	PcOpCode op = PC_OP_CONSTANT;
	int is = 0;
	switch(byte)
	{
	case 0:   op = PC_OP_END_INPUT;	  is=1; break;	// end of input
	case '(': op = PC_OP_OPEN_PAREN;  is=1; break;	// control
	case ')': op = PC_OP_CLOSE_PAREN; is=1; break;
	case ',': op = PC_OP_COMMA;       is=1; break;
	case '+': op = PC_OP_ADD;		  is=1; break;	// operators
	case '-': op = PC_OP_SUB;		  is=1; break;
	case '*': op = PC_OP_MULT;		  is=1; break;
	case '/': op = PC_OP_DIV;		  is=1; break;
	case '^': op = PC_OP_POW;		  is=1; break;
	} // switch
	if (is && pCode) *pCode = op;
	return(is);
}

// --------------------------------------------------------------
// determine if byte is white space
// returns: 0=no, 1=yes
int PlotCalc::IsWhiteSpace(char byte)
{
	if (byte==' ' || byte=='\n' || byte=='\r' || byte=='\t') return(1);
	return(0);
}

// --------------------------------------------------------------
// calculate
// this routine needs to be efficient
double PlotCalc::Calc(double x,double y,double z)
{
	double dtemp;
	unsigned ip; // instruction pointer

	m_sp = 0;	// clear stack pointer
	m_err_flags = 0;	// clear error flags
	for (ip=0; ip<m_nInst; ip++)
	{
		switch ((unsigned)m_opcodes[ip])
		{
	     case PC_OP_CONSTANT:IncSp(); m_dstack[m_sp] = m_dconstants[m_operands[ip]];           break;
	     case PC_OP_ADD:   DecSp(); m_dstack[m_sp] = m_dstack[m_sp] + m_dstack[m_sp+1];        break;
	     case PC_OP_SUB:   DecSp(); m_dstack[m_sp] = m_dstack[m_sp] - m_dstack[m_sp+1];	       break;
	     case PC_OP_MULT:  DecSp(); m_dstack[m_sp] = m_dstack[m_sp] * m_dstack[m_sp+1];	       break;
		 case PC_OP_DIV:   DecSp();
						   if (m_dstack[m_sp+1]==0.0)
						   {
						  	 m_dstack[m_sp] = 1.e10;
						  	 m_err_flags |= PC_ERR_DIVIDE_BY_ZERO;
						  	 break;
						   }
						   m_dstack[m_sp] = m_dstack[m_sp]/m_dstack[m_sp+1];
						   break;
	     case PC_OP_POW:   DecSp(); m_dstack[m_sp] = pow(m_dstack[m_sp],m_dstack[m_sp+1]);     break;
	     case PC_OP_POW2:  m_dstack[m_sp] = m_dstack[m_sp]*m_dstack[m_sp];                     break; // optimization
	     case PC_OP_POW3:  m_dstack[m_sp] = m_dstack[m_sp]*m_dstack[m_sp]*m_dstack[m_sp];      break; // optimization
	     case PC_OP_POW4:  dtemp = m_dstack[m_sp]*m_dstack[m_sp]; m_dstack[m_sp]=dtemp*dtemp;  break; // optimization
	     case PC_OP_NEG:   m_dstack[m_sp] = -m_dstack[m_sp];                                   break; // optimization
	     case PC_OP_PI:	   IncSp(); m_dstack[m_sp] = 3.141592653589793238462643;               break;
	     case PC_OP_E:	   IncSp(); m_dstack[m_sp] = 2.718281828459045235360287;		       break;
		 case PC_OP_U:	// use X code
		 case PC_OP_X:	   IncSp(); m_dstack[m_sp] = x;                                        break;
		 case PC_OP_V:  // use Y code
		 case PC_OP_Y:	   IncSp(); m_dstack[m_sp] = y;                                        break;
	     case PC_OP_Z:     IncSp(); m_dstack[m_sp] = z;                                        break;
		 case PC_OP_C1:    IncSp(); m_dstack[m_sp] = m_c1;									   break;
		 case PC_OP_C2:    IncSp(); m_dstack[m_sp] = m_c2;									   break;
		 case PC_OP_C3:    IncSp(); m_dstack[m_sp] = m_c3;									   break;
		 case PC_OP_C4:    IncSp(); m_dstack[m_sp] = m_c4;									   break;
		 case PC_OP_C5:    IncSp(); m_dstack[m_sp] = m_c5;									   break;
		 case PC_OP_C6:    IncSp(); m_dstack[m_sp] = m_c6;									   break;
		 case PC_OP_C7:    IncSp(); m_dstack[m_sp] = m_c7;									   break;
		 case PC_OP_C8:    IncSp(); m_dstack[m_sp] = m_c8;									   break;
		 case PC_OP_SETC1: DecSp(); m_c1 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC2: DecSp(); m_c2 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC3: DecSp(); m_c3 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC4: DecSp(); m_c4 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC5: DecSp(); m_c5 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC6: DecSp(); m_c6 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC7: DecSp(); m_c7 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SETC8: DecSp(); m_c8 = m_dstack[m_sp+1];								   break;
		 case PC_OP_SQRT:  if (m_dstack[m_sp]<0.)
						   {
							   m_dstack[m_sp] = 0.;
							   m_err_flags |= PC_ERR_NEGATIVE_SQRT;
							   break;
						   }
						   m_dstack[m_sp] = sqrt(m_dstack[m_sp]);
						   break;
	     case PC_OP_ABS:   m_dstack[m_sp] = fabs(m_dstack[m_sp]);                              break;
	     case PC_OP_SIN:   m_dstack[m_sp] =  sin(m_dstack[m_sp]);                              break;
	     case PC_OP_COS:   m_dstack[m_sp] =  cos(m_dstack[m_sp]); 							   break;
	     case PC_OP_TAN:   m_dstack[m_sp] =  tan(m_dstack[m_sp]); 							   break;
	     case PC_OP_ASIN:  dtemp = m_dstack[m_sp];
						   if (dtemp<-1.)
							   { dtemp=-1.; m_err_flags |= PC_ERR_OUT_OF_RANGE; }
						   if (dtemp> 1.)
							   { dtemp= 1.; m_err_flags |= PC_ERR_OUT_OF_RANGE; }
						   m_dstack[m_sp] = asin(dtemp);
						   break;
	     case PC_OP_ACOS:  dtemp = m_dstack[m_sp];
						   if (dtemp<-1.)
							   { dtemp=-1.; m_err_flags |= PC_ERR_OUT_OF_RANGE; }
						   if (dtemp> 1.)
							   { dtemp= 1.; m_err_flags |= PC_ERR_OUT_OF_RANGE; }
						   m_dstack[m_sp] = acos(dtemp);
						   break;
	     case PC_OP_ATAN:  m_dstack[m_sp] = atan(m_dstack[m_sp]); 							   break;
	     case PC_OP_SINH:  m_dstack[m_sp] = sinh(m_dstack[m_sp]); 							   break;
	     case PC_OP_COSH:  m_dstack[m_sp] = cosh(m_dstack[m_sp]); 							   break;
	     case PC_OP_TANH:  m_dstack[m_sp] = tanh(m_dstack[m_sp]); 							   break;
	     case PC_OP_EXP:   m_dstack[m_sp] = exp(m_dstack[m_sp]); 							   break;
	     case PC_OP_LN:    if (m_dstack[m_sp]<=0.)
						   {
						   	   m_dstack[m_sp] = 0.;
						   	   m_err_flags |= PC_ERR_OUT_OF_RANGE;
							   break;
						   }
						   m_dstack[m_sp] = log(m_dstack[m_sp]);
						   break;
	     case PC_OP_LOG10: if (m_dstack[m_sp]<=0.)
						   {
						   	   m_dstack[m_sp] = 0.;
						   	   m_err_flags |= PC_ERR_OUT_OF_RANGE;
							   break;
						   }
						   m_dstack[m_sp] = log10(m_dstack[m_sp]);
						   break;
	     case PC_OP_CEIL:  m_dstack[m_sp] = ceil(m_dstack[m_sp]); 							   break;
	     case PC_OP_FLOOR: m_dstack[m_sp] = floor(m_dstack[m_sp]); 							   break;
		 case PC_OP_RAND:  IncSp(); m_dstack[m_sp] = rand();								   break;
		 case PC_OP_MIN:   DecSp(); m_dstack[m_sp] = __min(m_dstack[m_sp],m_dstack[m_sp+1]);   break;
		 case PC_OP_MAX:   DecSp(); m_dstack[m_sp] = __max(m_dstack[m_sp],m_dstack[m_sp+1]);   break;
		 case PC_OP_ATAN2: DecSp(); m_dstack[m_sp] = atan2(m_dstack[m_sp],m_dstack[m_sp+1]);   break;
		 case PC_OP_HYPOT: DecSp(); m_dstack[m_sp] = hypot(m_dstack[m_sp],m_dstack[m_sp+1]);   break;
		 case PC_OP_SIGN:  m_dstack[m_sp] = (m_dstack[m_sp]>=0.) ? 1.0 : -1.0;                 break;

         // ### new functions
         case PC_OP_FACTORIAL: m_dstack[m_sp] = Factorial((int)m_dstack[m_sp]);                break;

         case PC_OP_IF_LT: DecSp(); DecSp(); DecSp();
                           if (m_dstack[m_sp] < m_dstack[m_sp+1])
                               m_dstack[m_sp] = m_dstack[m_sp+2];
                           else
                               m_dstack[m_sp] = m_dstack[m_sp+3];
                           break;
         case PC_OP_IF_LE: DecSp(); DecSp(); DecSp();
                           if (m_dstack[m_sp]<= m_dstack[m_sp+1])
                               m_dstack[m_sp] = m_dstack[m_sp+2];
                           else
                               m_dstack[m_sp] = m_dstack[m_sp+3];
                           break;
         case PC_OP_IF_GT: DecSp(); DecSp(); DecSp();
                           if (m_dstack[m_sp] > m_dstack[m_sp+1])
                               m_dstack[m_sp] = m_dstack[m_sp+2];
                           else
                               m_dstack[m_sp] = m_dstack[m_sp+3];
                           break;
         case PC_OP_IF_GE: DecSp(); DecSp(); DecSp();
                           if (m_dstack[m_sp]>= m_dstack[m_sp+1])
                               m_dstack[m_sp] = m_dstack[m_sp+2];
                           else
                               m_dstack[m_sp] = m_dstack[m_sp+3];
                           break;
		 case PC_OP_FMOD:  DecSp(); m_dstack[m_sp] = fmod(m_dstack[m_sp],m_dstack[m_sp+1]);    break;
	     case PC_OP_ASINH: dtemp = m_dstack[m_sp]; 
                           m_dstack[m_sp] = log(dtemp+sqrt(dtemp*dtemp+1.));                   
                           break;
	     case PC_OP_ACOSH: dtemp = m_dstack[m_sp];
                           if (dtemp<1.)
						   {
						   	   m_dstack[m_sp] = 0.;
						   	   m_err_flags |= PC_ERR_OUT_OF_RANGE;
							   break;
						   }
                           m_dstack[m_sp] = log(dtemp+sqrt(dtemp*dtemp-1.));                   
                           break;
	     case PC_OP_ATANH: dtemp = m_dstack[m_sp]; 
                           if (dtemp<= -1.) dtemp = -1. + 1.e-20;
                           if (dtemp>= +1.) dtemp = +1. - 1.e-20;
                           m_dstack[m_sp] = log((1.+dtemp)/(1.-dtemp))/2.;
                           break;

         case PC_OP_J0:    m_dstack[m_sp] =  j0(m_dstack[m_sp]);                               break;
         case PC_OP_J1:    m_dstack[m_sp] =  j1(m_dstack[m_sp]);                               break;
         case PC_OP_JN:    DecSp(); m_dstack[m_sp] = jn((int)m_dstack[m_sp],m_dstack[m_sp+1]); break;
         case PC_OP_Y0:    m_dstack[m_sp] =  y0(m_dstack[m_sp]);                               break;
         case PC_OP_Y1:    m_dstack[m_sp] =  y1(m_dstack[m_sp]);                               break;
         case PC_OP_YN:    DecSp(); m_dstack[m_sp] = yn((int)m_dstack[m_sp],m_dstack[m_sp+1]); break;

         case PC_OP_GAMMA: m_dstack[m_sp] = Gamma(m_dstack[m_sp]);                             break;
         case PC_OP_LGAMMA:m_dstack[m_sp] = LogGamma(m_dstack[m_sp]);                          break;
         case PC_OP_PSI:   m_dstack[m_sp] = Psi(m_dstack[m_sp]);                               break;
		 case PC_OP_ZETA:  DecSp(); m_dstack[m_sp] = Zetax(m_dstack[m_sp],m_dstack[m_sp+1]);   break;
		 case PC_OP_ZETAR: DecSp(); m_dstack[m_sp] = Zetar(m_dstack[m_sp],m_dstack[m_sp+1]);   break;
		 case PC_OP_ZETAI: DecSp(); m_dstack[m_sp] = Zetai(m_dstack[m_sp],m_dstack[m_sp+1]);   break;

		} // switch
	} // for
	return(m_dstack[1]);
}

// Factorial table
static long double const g_factorials[] = 
{ 
    1.000000000000000000000e+0L,   // 0! 
    1.000000000000000000000e+0L,   // 1! 
    2.000000000000000000000e+0L,   // 2! 
    6.000000000000000000000e+0L,   // 3! 
    2.400000000000000000000e+1L,   // 4! 
    1.200000000000000000000e+2L,   // 5! 
    7.200000000000000000000e+2L,   // 6! 
    5.040000000000000000000e+3L,   // 7! 
    4.032000000000000000000e+4L,   // 8! 
    3.628800000000000000000e+5L,   // 9! 
    3.628800000000000000000e+6L,   // 10! 
    3.991680000000000000000e+7L,   // 11! 
    4.790016000000000000000e+8L,   // 12! 
    6.227020800000000000000e+9L,   // 13! 
    8.717829120000000000000e+10L,  // 14! 
    1.307674368000000000000e+12L,  // 15! 
    2.092278988800000000000e+13L,  // 16! 
    3.556874280960000000000e+14L,  // 17! 
    6.402373705728000000000e+15L,  // 18! 
    1.216451004088320000000e+17L,  // 19! 
    2.432902008176640000000e+18L,  // 20! 
    5.109094217170944000000e+19L,  // 21! 
    1.124000727777607680000e+21L,  // 22! 
    2.585201673888497664000e+22L,  // 23! 
    6.204484017332394393600e+23L,  // 24! 
    1.551121004333098598400e+25L,  // 25! 
    4.032914611266056355840e+26L,  // 26! 
    1.088886945041835216077e+28L,  // 27! 
    3.048883446117138605015e+29L,  // 28! 
    8.841761993739701954544e+30L,  // 29! 
    2.652528598121910586363e+32L,  // 30! 
    8.222838654177922817726e+33L,  // 31! 
    2.631308369336935301672e+35L,  // 32! 
    8.683317618811886495518e+36L,  // 33! 
    2.952327990396041408476e+38L,  // 34! 
    1.033314796638614492967e+40L,  // 35! 
    3.719933267899012174680e+41L,  // 36! 
    1.376375309122634504632e+43L,  // 37! 
    5.230226174666011117600e+44L,  // 38! 
    2.039788208119744335864e+46L,  // 39! 
    8.159152832478977343456e+47L,  // 40! 
    3.345252661316380710817e+49L,  // 41! 
    1.405006117752879898543e+51L,  // 42! 
    6.041526306337383563736e+52L,  // 43! 
    2.658271574788448768044e+54L,  // 44! 
    1.196222208654801945620e+56L,  // 45! 
    5.502622159812088949850e+57L,  // 46! 
    2.586232415111681806430e+59L,  // 47! 
    1.241391559253607267086e+61L,  // 48! 
    6.082818640342675608723e+62L,  // 49! 
    3.041409320171337804361e+64L,  // 50! 
    1.551118753287382280224e+66L,  // 51! 
    8.065817517094387857166e+67L,  // 52! 
    4.274883284060025564298e+69L,  // 53! 
    2.308436973392413804721e+71L,  // 54! 
    1.269640335365827592597e+73L,  // 55! 
    7.109985878048634518540e+74L,  // 56! 
    4.052691950487721675568e+76L,  // 57! 
    2.350561331282878571829e+78L,  // 58! 
    1.386831185456898357379e+80L,  // 59! 
    8.320987112741390144276e+81L,  // 60! 
    5.075802138772247988009e+83L,  // 61! 
    3.146997326038793752565e+85L,  // 62! 
    1.982608315404440064116e+87L,  // 63! 
    1.268869321858841641034e+89L,  // 64! 
    8.247650592082470666723e+90L,  // 65! 
    5.443449390774430640037e+92L,  // 66! 
    3.647111091818868528825e+94L,  // 67! 
    2.480035542436830599601e+96L,  // 68! 
    1.711224524281413113725e+98L,  // 69! 
    1.197857166996989179607e+100L, // 70! 
    8.504785885678623175212e+101L, // 71! 
    6.123445837688608686152e+103L, // 72! 
    4.470115461512684340891e+105L, // 73! 
    3.307885441519386412260e+107L, // 74! 
    2.480914081139539809195e+109L, // 75! 
    1.885494701666050254988e+111L, // 76! 
    1.451830920282858696341e+113L, // 77! 
    1.132428117820629783146e+115L, // 78! 
    8.946182130782975286851e+116L, // 79! 
    7.156945704626380229481e+118L, // 80! 
    5.797126020747367985880e+120L, // 81! 
    4.753643337012841748421e+122L, // 82! 
    3.945523969720658651190e+124L, // 83! 
    3.314240134565353266999e+126L, // 84! 
    2.817104114380550276949e+128L, // 85! 
    2.422709538367273238177e+130L, // 86! 
    2.107757298379527717214e+132L, // 87! 
    1.854826422573984391148e+134L, // 88! 
    1.650795516090846108122e+136L, // 89! 
    1.485715964481761497310e+138L, // 90! 
    1.352001527678402962552e+140L, // 91! 
    1.243841405464130725548e+142L, // 92! 
    1.156772507081641574759e+144L, // 93! 
    1.087366156656743080274e+146L, // 94! 
    1.032997848823905926260e+148L, // 95! 
    9.916779348709496892096e+149L, // 96! 
    9.619275968248211985333e+151L, // 97! 
    9.426890448883247745626e+153L, // 98! 
    9.332621544394415268170e+155L, // 99! 
    9.332621544394415268170e+157L, // 100! 
    9.425947759838359420852e+159L, // 101! 
    9.614466715035126609269e+161L, // 102! 
    9.902900716486180407547e+163L, // 103! 
    1.029901674514562762385e+166L, // 104! 
    1.081396758240290900504e+168L, // 105! 
    1.146280563734708354534e+170L, // 106! 
    1.226520203196137939352e+172L, // 107! 
    1.324641819451828974500e+174L, // 108! 
    1.443859583202493582205e+176L, // 109! 
    1.588245541522742940425e+178L, // 110! 
    1.762952551090244663872e+180L, // 111!
    1.974506857221074023537e+182L, // 112! 
    2.231192748659813646597e+184L, // 113! 
    2.543559733472187557120e+186L, // 114! 
    2.925093693493015690688e+188L, // 115! 
    3.393108684451898201198e+190L, // 116! 
    3.969937160808720895402e+192L, // 117! 
    4.684525849754290656574e+194L, // 118! 
    5.574585761207605881323e+196L, // 119!
    6.689502913449127057588e+198L, // 120! 
    8.094298525273443739682e+200L, // 121! 
    9.875044200833601362412e+202L, // 122! 
    1.214630436702532967577e+205L, // 123! 
    1.506141741511140879795e+207L, // 124! 
    1.882677176888926099744e+209L, // 125! 
    2.372173242880046885677e+211L, // 126! 
    3.012660018457659544810e+213L, // 127! 
    3.856204823625804217357e+215L, // 128! 
    4.974504222477287440390e+217L, // 129! 
    6.466855489220473672507e+219L, // 130! 
    8.471580690878820510985e+221L, // 131! 
    1.118248651196004307450e+224L, // 132! 
    1.487270706090685728908e+226L, // 133! 
    1.992942746161518876737e+228L, // 134! 
    2.690472707318050483595e+230L, // 135! 
    3.659042881952548657690e+232L, // 136! 
    5.012888748274991661035e+234L, // 137! 
    6.917786472619488492228e+236L, // 138! 
    9.615723196941089004197e+238L, // 139! 
    1.346201247571752460588e+241L, // 140! 
    1.898143759076170969429e+243L, // 141! 
    2.695364137888162776589e+245L, // 142! 
    3.854370717180072770522e+247L, // 143! 
    5.550293832739304789551e+249L, // 144! 
    8.047926057471991944849e+251L, // 145! 
    1.174997204390910823948e+254L, // 146! 
    1.727245890454638911203e+256L, // 147! 
    2.556323917872865588581e+258L, // 148! 
    3.808922637630569726986e+260L, // 149! 
    5.713383956445854590479e+262L, // 150! 
    8.627209774233240431623e+264L, // 151! 
    1.311335885683452545607e+267L, // 152! 
    2.006343905095682394778e+269L, // 153! 
    3.089769613847350887959e+271L, // 154! 
    4.789142901463393876336e+273L, // 155! 
    7.471062926282894447084e+275L, // 156! 
    1.172956879426414428192e+278L, // 157! 
    1.853271869493734796544e+280L, // 158! 
    2.946702272495038326504e+282L, // 159! 
    4.714723635992061322407e+284L, // 160! 
    7.590705053947218729075e+286L, // 161! 
    1.229694218739449434110e+289L, // 162! 
    2.004401576545302577600e+291L, // 163! 
    3.287218585534296227263e+293L, // 164! 
    5.423910666131588774984e+295L, // 165! 
    9.003691705778437366474e+297L, // 166! 
    1.503616514864999040201e+300L, // 167! 
    2.526075744973198387538e+302L, // 168! 
    4.269068009004705274939e+304L, // 169! 
    7.257415615307998967397e+306L  // 170! 
};

static const int NFACTORIALS = sizeof(g_factorials) / sizeof(g_factorials[0]);

// --------------------------------------------------------------------
double Factorial(int n)
{ 
    if ( n < 0 ) return 0.0; 
    if ( n >= NFACTORIALS ) return DBL_MAX; 
    return (double) g_factorials[n]; 
} 

// --------------------------------------------------------------------
//  gamma.cpp -- computation of gamma function.
//      Algorithms and coefficient values from "Computation of Special
//      Functions", Zhang and Jin, John Wiley and Sons, 1996.
//  (C) 2003, C. Bond. All rights reserved.
//
// Returns gamma function of argument 'x'.
//
// NOTE: Returns 1e308 if argument is a negative integer or 0,
//       or if argument exceeds 171.
//
// --------------------------------------------------------------------
double Gamma(double x)
{
    int k,m;
    double ga,gr,r=1.,z;

    static double g[] = {
        1.0,
        0.5772156649015329,
       -0.6558780715202538,
       -0.420026350340952e-1,
        0.1665386113822915,
       -0.421977345555443e-1,
       -0.9621971527877e-2,
        0.7218943246663e-2,
       -0.11651675918591e-2,
       -0.2152416741149e-3,
        0.1280502823882e-3,
       -0.201348547807e-4,
       -0.12504934821e-5,
        0.1133027232e-5,
       -0.2056338417e-6,
        0.6116095e-8,
        0.50020075e-8,
       -0.11812746e-8,
        0.1043427e-9,
        0.77823e-11,
       -0.36968e-11,
        0.51e-12,
       -0.206e-13,
       -0.54e-14,
        0.14e-14};

    if (x > 171.0) return DBL_MAX;    // This value is an overflow flag.
    if (x == (int)x) 
    {
        if (x > 0.0) 
        {
            /*
            ga = 1.0;               // use factorial
            for (i=2;i<x;i++) 
            {
               ga *= i;
            }
            */
            ga = Factorial((int)x);
         }
         else
            ga = DBL_MAX;
     }
     else 
     {
        if (fabs(x) > 1.0) 
        {
            z = fabs(x);
            m = (int)z;
            r = 1.0;
            for (k=1;k<=m;k++) 
            {
                r *= (z-k);
            }
            z -= m;
        }
        else
            z = x;
        gr = g[24];
        for (k=23;k>=0;k--) 
        {
            gr = gr*z+g[k];
        }
        ga = 1.0/(gr*z);
        if (fabs(x) > 1.0) 
        {
            ga *= r;
            if (x < 0.0) 
            {
                ga = -M_PI/(x*ga*sin(M_PI*x));
            }
        }
    }
    return(ga);
}

// --------------------------------------------------------------------
// log gamma function of real argument.
//      Algorithms and coefficient values from "Computation of Special
//      Functions", Zhang and Jin, John Wiley and Sons, 1996.
//
//  (C) 2003, C. Bond. All rights reserved.
//
//  Returns log(gamma) of real argument.
//  NOTE: Returns 1e308 if argument is 0 or negative.
// --------------------------------------------------------------------
//
#include <math.h>
double LogGamma(double x)
{
    double x0,x2,xp,gl,gl0;
    int n=0,k;
    static double a[] = 
    {
        8.333333333333333e-02,
       -2.777777777777778e-03,
        7.936507936507937e-04,
       -5.952380952380952e-04,
        8.417508417508418e-04,
       -1.917526917526918e-03,
        6.410256410256410e-03,
       -2.955065359477124e-02,
        1.796443723688307e-01,
       -1.39243221690590
    };
    
    x0 = x;
    if (x <= 0.0) return DBL_MAX;
    else if ((x == 1.0) || (x == 2.0)) return 0.0;
    else if (x <= 7.0) 
    {
        n = (int)(7-x);
        x0 = x+n;
    }
    x2 = 1.0/(x0*x0);
    xp = 2.0*M_PI;
    gl0 = a[9];
    for (k=8;k>=0;k--) 
    {
        gl0 = gl0*x2 + a[k];
    }
    gl = gl0/x0+0.5*log(xp)+(x0-0.5)*log(x0)-x0;
    if (x <= 7.0) 
    {
        for (k=1;k<=n;k++) 
        {
            gl -= log(x0-1.0);
            x0 -= 1.0;
        }
    }
    return gl;
}

// --------------------------------------------------------------------
// psi function for real arguments.
//      Algorithms and coefficient values from "Computation of Special
//      Functions", Zhang and Jin, John Wiley and Sons, 1996.
//
//  (C) 2003, C. Bond. All rights reserved.
//
//  Returns psi function for real argument 'x'.
//  NOTE: Returns 1e308 for argument 0 or a negative integer.
//
// --------------------------------------------------------------------
#define EL  (0.5772156649015329)

double Psi(double x)
{
    double s,ps,xa,x2;
    int n,k;
    static double a[] = {
        -0.8333333333333e-01,
         0.83333333333333333e-02,
        -0.39682539682539683e-02,
         0.41666666666666667e-02,
        -0.75757575757575758e-02,
         0.21092796092796093e-01,
        -0.83333333333333333e-01,
         0.4432598039215686};

    xa = fabs(x);
    s = 0.0;
    if ((x == (int)x) && (x <= 0.0)) 
    {
        ps = DBL_MAX;
        return ps;
    }
    if (xa == (int)xa) 
    {
        n = (int)xa;
        for (k=1;k<n;k++) 
        {
            s += 1.0/k;
        }
        ps =  s-EL;
    }
    else if ((xa+0.5) == ((int)(xa+0.5))) 
    {
        n = (int)(xa-0.5);
        for (k=1;k<=n;k++) 
        {
            s += 1.0/(2.0*k-1.0);
        }
        ps = 2.0*s-EL-1.386294361119891;
    }
    else 
    {
        if (xa < 10.0) 
        {
            n = 10-(int)xa;
            for (k=0;k<n;k++) 
            {
                s += 1.0/(xa+k);
            }
            xa += n;
        }
        x2 = 1.0/(xa*xa);
        ps = log(xa)-0.5/xa+x2*(((((((a[7]*x2+a[6])*x2+a[5])*x2+
            a[4])*x2+a[3])*x2+a[2])*x2+a[1])*x2+a[0]);
        ps -= s;
    }
    if (x < 0.0)
        ps = ps - M_PI*cos(M_PI*x)/sin(M_PI*x)-1.0/x;
    return ps;
}

// --------------------------------------------------------------
// Zeta Function
#define NLIMIT (1000) 
void Zeta(double real, double imag, double* rout, double* iout, unsigned* count)
{
    double zreal=1, zimag=0;
    double logn, c1, c2, c3, denom, amp, phase, amp2=0, phase2=0, diff, total;
    unsigned int i;

    /* 
    zeta(z)= fE(z)*zeta(1-z)
    with fE(z)= Gamma(1-z)*(2*pi)^(z-1)*2*sin(z*pi/2) 
    */

    for (i=2; i<NLIMIT; )
    {
        // even item
        logn   = log(i++);
		c1     = -real*logn;
        c2     = -imag*logn;
		c3     = exp(c1);
        zreal -= c3*cos(c2);
        zimag -= c3*sin(c2);

        // odd item
        logn   = log(i++);
		c1     = -real*logn;
        c2     = -imag*logn;
		c3     = exp(c1);
        zreal += c3*cos(c2);
        zimag += c3*sin(c2);

        // determine convergence
        amp   = sqrt(zreal*zreal + zimag*zimag);
        phase = atan2(zimag, zreal);

        diff  = fabs(fabs(amp) - fabs(amp2));
        total = fabs(amp);
        if (diff/total < 1.e-12)
            break;

        // remember as last
        amp2   = amp;
        phase2 = phase;
    } // for
    denom = 1. - pow(2, 1.-real);
    // return values if requested
	if (rout ) *rout = zreal/denom;
	if (iout ) *iout = zimag/denom;
    if (count) *count = i;
}

// returns real part of zeta function
double Zetar(double real, double imag)
{
    double value;
    Zeta(real, imag, &value, NULL, NULL);
    return(value);
}

// returns imaginary part of zeta function
double Zetai(double real, double imag)
{
    double value;
    Zeta(real, imag, NULL, &value, NULL);
    return(value);
}

// returns amplitude of zeta function
double Zetax(double real, double imag)
{
    double value, zreal, zimag;
    Zeta(real, imag, &zreal, &zimag, NULL);
    value = sqrt(zreal*zreal+zimag*zimag);
    return(value);
}

/*	<><><><><><><><><><><><><>  PlotCalc.cpp  <><><><><><><><><><><><><><> */
